import Color from 'color';
interface StateRule {
    metadata: {
        x: number;
        y: number;
    };
    events: Record<string, string>;
}
interface StateMachine {
    initialState: string;
    rules: Record<string, StateRule>;
}
interface TransitionTarget {
    name: string;
    x: number;
    y: number;
}
interface Transition {
    src: TransitionTarget;
    dest: TransitionTarget;
    event: string;
    path: any;
    pointer: any;
}
interface EditorOptions {
    stateRadius?: number;
    stateFill?: Color;
    stateStroke?: Color;
    initialStateStroke?: Color;
    textStroke?: Color;
    backgroundColor?: Color;
    selectionColor?: Color;
    selfLinkHeight?: number;
    selectableStates?: boolean;
    maxZoom?: number;
    minZoom?: number;
}
export default class StateMachineEditor extends HTMLElement {
    #private;
    stateMachineRules: Record<string, StateRule>;
    initialState: string;
    transitions: any[];
    states: any[];
    statesMap: Record<string, any>;
    constructor();
    connectedCallback(): void;
    get options(): EditorOptions;
    set options(newOptions: EditorOptions);
    get stateMachine(): StateMachine;
    set stateMachine(value: StateMachine);
    get width(): string | null;
    set width(newWidth: string | null);
    get height(): string | null;
    set height(newHeight: string | null);
    static get observedAttributes(): string[];
    attributeChangedCallback(name: string, _oldValue: string, newValue: string): void;
    private initialRender;
    setStateMachine({ initialState, rules }: StateMachine): void;
    addState(name: string, x: number, y: number): void;
    setInitialState(name: string): void;
    deleteState(name: string): void;
    deleteTransition({ event, src, dest, path, pointer }: Transition): void;
    startAddTransition(event: string, srcName: string, destX: number, destY: number): void;
    stopAddTransition(): void;
    setCanvasSize(width: number, height: number): void;
    setHeight(height: number): void;
    setWidth(width: number): void;
    setHoverCursor(cursor: string): void;
    getStateMachine(): any;
    private parseObjectName;
    private dispatchCustomEvent;
    private updateAllObjects;
    private handleMouseWheel;
    private handleMouseDown;
    private handleMouseUp;
    private updateTransition;
    private handleMouseMove;
    private handleObjectMove;
    private handleDoubleClick;
    private handleStateMachineChange;
    private handleStateDoubleClick;
    private handleStateMouseOver;
    private handleStateMouseOut;
    private handleStateClick;
    private handleTransitionClick;
    private handleTransitionOver;
    private handleTransitionOut;
    private drawTransition;
    private drawState;
    private calcLinkPath;
    private calcPointerPath;
    private calcSelfLinkPath;
    private calcTransitionLink;
}
export {};
