"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to set private field on non-instance");
    }
    privateMap.set(receiver, value);
    return value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return privateMap.get(receiver);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _root, _canvasElement, _canvas, _isDragging, _newTransition, _lastPosX, _lastPosY, _selectableStates, _stateRadius, _stateFill, _stateStroke, _initialStateStroke, _textStroke, _backgroundColor, _selectionColor, _selfLinkHeight, _maxZoom, _minZoom;
Object.defineProperty(exports, "__esModule", { value: true });
const fabric_1 = require("fabric");
const color_hash_1 = __importDefault(require("color-hash"));
const color_1 = __importDefault(require("color"));
const colorHash = new color_hash_1.default();
const ELEMENT_EVENTS = {
    stateMachineChange: 'statemachine:change',
    state: {
        click: 'state:click',
        rightClick: 'state:rightclick',
        over: 'state:over',
        out: 'state:out',
        doubleClick: 'state:doubleclick',
    },
    transition: {
        click: 'transition:click',
        rightClick: 'transition:rightclick',
        over: 'transition:over',
        out: 'transition:out',
        doubleClick: 'transition:doubleclick',
    },
    canvas: {
        click: 'canvas:click',
        rightClick: 'canvas:rightclick',
        doubleClick: 'canvas:doubleclick',
    },
};
const EDITOR_DEFAULTS = {
    stateFill: new color_1.default('#180019'),
    stateStroke: new color_1.default('#C5CBD3'),
    initialStateStroke: new color_1.default('orange'),
    textStroke: new color_1.default('white'),
    backgroundColor: new color_1.default('#330031'),
    selectionColor: new color_1.default(),
    selfLinkHeight: 100,
    stateRadius: 30,
    maxZoom: 20,
    minZoom: 0.1,
    selectableStates: true,
};
class StateMachineEditor extends HTMLElement {
    constructor() {
        super();
        // html and canvas elements
        _root.set(this, void 0);
        _canvasElement.set(this, void 0);
        _canvas.set(this, void 0);
        // internal state
        _isDragging.set(this, false);
        _newTransition.set(this, null);
        _lastPosX.set(this, 0);
        _lastPosY.set(this, 0);
        this.stateMachineRules = {};
        this.initialState = '';
        this.transitions = [];
        this.states = [];
        this.statesMap = {};
        // configuration
        _selectableStates.set(this, EDITOR_DEFAULTS.selectableStates);
        _stateRadius.set(this, EDITOR_DEFAULTS.stateRadius);
        _stateFill.set(this, EDITOR_DEFAULTS.stateFill);
        _stateStroke.set(this, EDITOR_DEFAULTS.stateStroke);
        _initialStateStroke.set(this, EDITOR_DEFAULTS.initialStateStroke);
        _textStroke.set(this, EDITOR_DEFAULTS.textStroke);
        _backgroundColor.set(this, EDITOR_DEFAULTS.backgroundColor);
        _selectionColor.set(this, EDITOR_DEFAULTS.selectionColor);
        _selfLinkHeight.set(this, EDITOR_DEFAULTS.selfLinkHeight);
        _maxZoom.set(this, EDITOR_DEFAULTS.maxZoom);
        _minZoom.set(this, EDITOR_DEFAULTS.minZoom);
        __classPrivateFieldSet(this, _root, this.attachShadow({ mode: 'open' }));
        __classPrivateFieldSet(this, _canvasElement, document.createElement('canvas'));
        __classPrivateFieldGet(this, _root).appendChild(__classPrivateFieldGet(this, _canvasElement));
        __classPrivateFieldSet(this, _canvas, __classPrivateFieldSet(this, _canvas, new fabric_1.fabric.Canvas(__classPrivateFieldGet(this, _canvasElement), {
            backgroundColor: __classPrivateFieldGet(this, _backgroundColor).toString(),
            selection: false,
            fireRightClick: true,
            stopContextMenu: true,
        })));
        __classPrivateFieldGet(this, _canvas).on('mouse:down', this.handleMouseDown.bind(this));
        __classPrivateFieldGet(this, _canvas).on('mouse:up', this.handleMouseUp.bind(this));
        __classPrivateFieldGet(this, _canvas).on('mouse:wheel', this.handleMouseWheel.bind(this));
        __classPrivateFieldGet(this, _canvas).on('mouse:move', this.handleMouseMove.bind(this));
        __classPrivateFieldGet(this, _canvas).on('object:moving', this.handleObjectMove.bind(this));
        __classPrivateFieldGet(this, _canvas).on('mouse:dblclick', this.handleDoubleClick.bind(this));
    }
    connectedCallback() {
        this.setCanvasSize(Number(this.getAttribute('width')) || 100, Number(this.getAttribute('height')) || 100);
    }
    get options() {
        return {
            selectableStates: __classPrivateFieldGet(this, _selectableStates),
            stateRadius: __classPrivateFieldGet(this, _stateRadius),
            stateFill: __classPrivateFieldGet(this, _stateFill),
            stateStroke: __classPrivateFieldGet(this, _stateStroke),
            initialStateStroke: __classPrivateFieldGet(this, _initialStateStroke),
            textStroke: __classPrivateFieldGet(this, _textStroke),
            backgroundColor: __classPrivateFieldGet(this, _backgroundColor),
            selectionColor: __classPrivateFieldGet(this, _selectionColor),
            selfLinkHeight: __classPrivateFieldGet(this, _selfLinkHeight),
        };
    }
    set options(newOptions) {
        __classPrivateFieldSet(this, _selectableStates, newOptions.selectableStates || __classPrivateFieldGet(this, _selectableStates));
        __classPrivateFieldSet(this, _stateRadius, newOptions.stateRadius || __classPrivateFieldGet(this, _stateRadius));
        __classPrivateFieldSet(this, _stateFill, newOptions.stateFill || __classPrivateFieldGet(this, _stateFill));
        __classPrivateFieldSet(this, _stateStroke, newOptions.stateStroke || __classPrivateFieldGet(this, _stateStroke));
        __classPrivateFieldSet(this, _initialStateStroke, newOptions.initialStateStroke || __classPrivateFieldGet(this, _initialStateStroke));
        __classPrivateFieldSet(this, _textStroke, newOptions.textStroke || __classPrivateFieldGet(this, _textStroke));
        __classPrivateFieldSet(this, _backgroundColor, newOptions.backgroundColor || __classPrivateFieldGet(this, _backgroundColor));
        __classPrivateFieldSet(this, _selectionColor, newOptions.selectionColor || __classPrivateFieldGet(this, _selectionColor));
        __classPrivateFieldSet(this, _selfLinkHeight, newOptions.selfLinkHeight || __classPrivateFieldGet(this, _selfLinkHeight));
        __classPrivateFieldSet(this, _maxZoom, newOptions.maxZoom || __classPrivateFieldGet(this, _maxZoom));
        __classPrivateFieldSet(this, _minZoom, newOptions.minZoom || __classPrivateFieldGet(this, _minZoom));
        this.initialRender();
    }
    get stateMachine() {
        return this.getStateMachine();
    }
    set stateMachine(value) {
        this.setStateMachine(value);
    }
    get width() {
        return this.getAttribute('width');
    }
    set width(newWidth) {
        if (newWidth == null) {
            return;
        }
        this.setWidth(Number(newWidth));
        this.setAttribute('width', newWidth);
    }
    get height() {
        return this.getAttribute('height');
    }
    set height(newHeight) {
        if (newHeight == null)
            return;
        this.setHeight(Number(newHeight));
        this.setAttribute('width', newHeight);
    }
    static get observedAttributes() {
        return ['width', 'height'];
    }
    attributeChangedCallback(name, _oldValue, newValue) {
        switch (name) {
            case 'width':
                this.setWidth(Number(newValue));
                break;
            case 'height':
                this.setHeight(Number(newValue));
                break;
            default:
                console.log('defalut case');
                break;
        }
    }
    initialRender() {
        // clear old render
        this.transitions = [];
        this.states = [];
        this.statesMap = {};
        __classPrivateFieldGet(this, _canvas).clear();
        __classPrivateFieldGet(this, _canvas).setBackgroundColor(__classPrivateFieldGet(this, _backgroundColor).toString(), () => { });
        // new render
        const stateList = Object.keys(this.stateMachineRules);
        stateList.forEach((stateName) => {
            const { metadata: { x, y }, events, } = this.stateMachineRules[stateName];
            this.drawState(stateName, x, y);
            const eventKeys = Object.keys(events);
            eventKeys.forEach((event) => {
                const destStateName = events[event];
                const dest = this.stateMachineRules[destStateName].metadata;
                const { path, pointer } = this.drawTransition(x, y, dest.x, dest.y, event);
                this.transitions.push({
                    event,
                    path,
                    pointer,
                    dest: { name: destStateName, ...dest },
                    src: { name: stateName, x, y },
                });
            });
        });
    }
    // ACTIONS
    setStateMachine({ initialState, rules }) {
        this.stateMachineRules = rules;
        this.initialState = initialState;
        this.initialRender();
    }
    addState(name, x, y) {
        if (typeof name !== 'string' || this.stateMachineRules[name] != null) {
            throw new Error('State Already Exists');
        }
        this.stateMachineRules[name] = {
            metadata: { x, y },
            events: {},
        };
        this.drawState(name, x, y);
        this.handleStateMachineChange();
    }
    setInitialState(name) {
        const newInitialState = this.statesMap[name];
        const oldInitialState = this.statesMap[this.initialState];
        if (newInitialState)
            newInitialState._objects[0].set('stroke', __classPrivateFieldGet(this, _initialStateStroke));
        if (oldInitialState)
            oldInitialState._objects[0].set('stroke', __classPrivateFieldGet(this, _stateStroke));
        this.initialState = name;
        __classPrivateFieldGet(this, _canvas).renderAll();
    }
    deleteState(name) {
        if (this.statesMap[name] == null) {
            throw new Error('Invalid State Item');
        }
        __classPrivateFieldGet(this, _canvas).remove(this.statesMap[name]);
        this.transitions = this.transitions.filter(({ src, dest }) => !(src.name !== name || dest.name === name));
        delete this.statesMap[name];
        delete this.stateMachineRules[name];
        const statemEntries = Object.entries(this.stateMachineRules).map(([key, value]) => {
            const values = Object.entries(value).filter((entry) => {
                return value !== entry[1];
            });
            return [key, Object.fromEntries(values)];
        });
        this.stateMachineRules = Object.fromEntries(statemEntries);
        this.handleStateMachineChange();
    }
    deleteTransition({ event, src, dest, path, pointer }) {
        this.transitions = this.transitions.filter((t) => {
            const flag = event === t.event && src === t.src && dest === t.dest;
            if (flag) {
                __classPrivateFieldGet(this, _canvas).remove(path);
                __classPrivateFieldGet(this, _canvas).remove(pointer);
                return false;
            }
            return true;
        });
        delete this.stateMachineRules[src.name].events[event];
        this.handleStateMachineChange();
    }
    startAddTransition(event, srcName, destX, destY) {
        if (this.statesMap[srcName] == null) {
            throw new Error('Invalid State Item');
        }
        if (__classPrivateFieldGet(this, _newTransition) != null) {
            throw new Error('Already adding transition');
        }
        const stateObject = this.statesMap[srcName];
        __classPrivateFieldSet(this, _newTransition, this.drawTransition(stateObject.left, stateObject.top, destX, destY, event));
        __classPrivateFieldGet(this, _newTransition).event = event;
        __classPrivateFieldGet(this, _newTransition).src = {
            name: srcName,
            x: stateObject.left,
            y: stateObject.top,
        };
        __classPrivateFieldGet(this, _newTransition).dest = null;
    }
    stopAddTransition() {
        if (__classPrivateFieldGet(this, _newTransition) == null)
            return;
        const { pointer, path } = __classPrivateFieldGet(this, _newTransition);
        __classPrivateFieldGet(this, _canvas).remove(pointer);
        __classPrivateFieldGet(this, _canvas).remove(path);
        __classPrivateFieldSet(this, _newTransition, null);
    }
    setCanvasSize(width, height) {
        __classPrivateFieldGet(this, _canvas).setWidth(width);
        __classPrivateFieldGet(this, _canvas).setHeight(height);
    }
    setHeight(height) {
        __classPrivateFieldGet(this, _canvas).setHeight(height);
    }
    setWidth(width) {
        __classPrivateFieldGet(this, _canvas).setWidth(width);
    }
    setHoverCursor(cursor) {
        __classPrivateFieldGet(this, _canvas).hoverCursor = cursor;
    }
    getStateMachine() {
        this.states.forEach((state) => {
            const result = this.parseObjectName(state.name);
            if (result) {
                this.stateMachineRules[result.name].metadata = {
                    x: state.left,
                    y: state.top,
                };
            }
        });
        return JSON.parse(JSON.stringify(this.stateMachineRules));
    }
    // UTILS
    parseObjectName(objectName) {
        if (objectName == null) {
            return null;
        }
        const [type, name] = objectName.split(':');
        return {
            type,
            name,
        };
    }
    // CANVAS EVENT HANDLERS
    dispatchCustomEvent(name, detail) {
        this.dispatchEvent(new CustomEvent(name, {
            detail,
        }));
    }
    updateAllObjects() {
        __classPrivateFieldGet(this, _canvas).requestRenderAll();
        this.states.forEach((state) => state.setCoords());
        this.transitions.forEach(({ pointer, path }) => {
            pointer.setCoords();
            path.setCoords();
        });
    }
    handleMouseWheel({ e }) {
        e.preventDefault();
        e.stopPropagation();
        const { offsetX, offsetY, deltaY } = e;
        let zoom = __classPrivateFieldGet(this, _canvas).getZoom() + deltaY / 200;
        if (zoom > __classPrivateFieldGet(this, _maxZoom))
            zoom = __classPrivateFieldGet(this, _maxZoom);
        if (zoom < __classPrivateFieldGet(this, _minZoom))
            zoom = __classPrivateFieldGet(this, _minZoom);
        __classPrivateFieldGet(this, _canvas).zoomToPoint(new fabric_1.fabric.Point(offsetX, offsetY), zoom);
        this.updateAllObjects();
    }
    handleMouseDown({ button, target, e, absolutePointer, }) {
        const { altKey, clientX, clientY } = e;
        if (target != null)
            return;
        if (altKey) {
            __classPrivateFieldSet(this, _isDragging, true);
            __classPrivateFieldSet(this, _lastPosX, clientX);
            __classPrivateFieldSet(this, _lastPosY, clientY);
            return;
        }
        const detail = {
            event: e,
            canvas: { x: absolutePointer?.x, y: absolutePointer?.y },
        };
        if (button === 3) {
            this.dispatchCustomEvent(ELEMENT_EVENTS.canvas.rightClick, detail);
        }
        else {
            this.dispatchCustomEvent(ELEMENT_EVENTS.canvas.click, detail);
        }
    }
    handleMouseUp() {
        __classPrivateFieldSet(this, _isDragging, false);
    }
    updateTransition(x, y) {
        if (__classPrivateFieldGet(this, _newTransition) == null)
            return;
        const { src, pointer, path, dest, event } = __classPrivateFieldGet(this, _newTransition);
        if (src && dest == null) {
            __classPrivateFieldGet(this, _canvas).remove(pointer);
            __classPrivateFieldGet(this, _canvas).remove(path);
            __classPrivateFieldSet(this, _newTransition, this.drawTransition(src.x, src.y, x, y, event));
            __classPrivateFieldGet(this, _newTransition).src = src;
            __classPrivateFieldGet(this, _newTransition).event = event;
            __classPrivateFieldGet(this, _newTransition).dest = null;
        }
    }
    handleMouseMove({ absolutePointer, e }) {
        if (absolutePointer && __classPrivateFieldGet(this, _isDragging)) {
            const { clientX, clientY } = e;
            const { x, y } = absolutePointer;
            this.updateTransition(x, y);
            if (__classPrivateFieldGet(this, _canvas).viewportTransform &&
                __classPrivateFieldGet(this, _canvas).viewportTransform[4] &&
                __classPrivateFieldGet(this, _canvas).viewportTransform[5]) {
                __classPrivateFieldGet(this, _canvas).viewportTransform[4] += clientX - __classPrivateFieldGet(this, _lastPosX);
                __classPrivateFieldGet(this, _canvas).viewportTransform[5] += clientY - __classPrivateFieldGet(this, _lastPosY);
            }
            this.updateAllObjects();
            __classPrivateFieldSet(this, _lastPosX, clientX);
            __classPrivateFieldSet(this, _lastPosY, clientY);
        }
    }
    handleObjectMove({ target: _target }) {
        const target = _target;
        if (target == null)
            return;
        const stateData = this.parseObjectName(target.name || '');
        if (stateData == null) {
            return;
        }
        const { type, name } = stateData;
        if (type === 'state') {
            this.transitions = this.transitions.map((link) => {
                const { dest, src } = link;
                if (dest.name !== name && src.name !== name) {
                    return link;
                }
                if (dest.name === name) {
                    dest.x = target.left;
                    dest.y = target.top;
                }
                if (src.name === name) {
                    src.x = target.left;
                    src.y = target.top;
                }
                __classPrivateFieldGet(this, _canvas).remove(link.path);
                __classPrivateFieldGet(this, _canvas).remove(link.pointer);
                const { path, pointer } = this.drawTransition(src.x, src.y, dest.x, dest.y, link.event);
                return { ...link, path, pointer };
            });
            this.handleStateMachineChange();
        }
    }
    handleDoubleClick({ absolutePointer, target: _target, e: event, }) {
        if (_target == null)
            return;
        const target = _target;
        const { name: targetName = '' } = target;
        const result = this.parseObjectName(targetName);
        if (result == null)
            return;
        const { name } = result;
        this.dispatchCustomEvent(ELEMENT_EVENTS.canvas.doubleClick, {
            state: { name, x: target.left, y: target.top },
            canvas: { x: absolutePointer?.x, y: absolutePointer?.y },
            event: event,
        });
    }
    // STATE EVENTS
    handleStateMachineChange() {
        this.dispatchCustomEvent(ELEMENT_EVENTS.stateMachineChange, {
            stateMachine: this.getStateMachine(),
        });
    }
    handleStateDoubleClick({ absolutePointer, target: _target, e: event, }) {
        if (_target == null)
            return;
        const target = _target;
        const { name: targetName = '' } = target;
        const result = this.parseObjectName(targetName);
        if (result == null)
            return;
        const { name } = result;
        this.dispatchCustomEvent(ELEMENT_EVENTS.state.doubleClick, {
            state: { name, x: target.left, y: target.top },
            canvas: { x: absolutePointer?.x, y: absolutePointer?.y },
            event: event,
        });
    }
    handleStateMouseOver({ target: _target, e: event }) {
        if (_target == null)
            return;
        const target = _target;
        const { name: targetName = '' } = target;
        const result = this.parseObjectName(targetName);
        if (result == null)
            return;
        const { name } = result;
        this.dispatchCustomEvent(ELEMENT_EVENTS.state.over, {
            state: { name, x: target.left, y: target.top },
            event: event,
        });
        if (__classPrivateFieldGet(this, _newTransition)) {
            const { src, path, pointer, event } = __classPrivateFieldGet(this, _newTransition);
            __classPrivateFieldGet(this, _canvas).remove(path);
            __classPrivateFieldGet(this, _canvas).remove(pointer);
            __classPrivateFieldSet(this, _newTransition, this.drawTransition(src.x, src.y, target.left || 0, target.top || 0, event));
            __classPrivateFieldGet(this, _newTransition).src = src;
            __classPrivateFieldGet(this, _newTransition).event = event;
            __classPrivateFieldGet(this, _newTransition).dest = {
                name,
                x: target.left,
                y: target.top,
            };
        }
        if (target._objects.length > 0) {
            target._objects[0].set('stroke', 'red');
            target._objects[0].animate('radius', 32, {
                duration: 300,
                onChange: __classPrivateFieldGet(this, _canvas).renderAll.bind(__classPrivateFieldGet(this, _canvas)),
            });
        }
    }
    handleStateMouseOut({ target: _target, e }) {
        if (_target == null)
            return;
        const target = _target;
        const { name: targetName = '' } = target;
        const result = this.parseObjectName(targetName);
        if (result == null)
            return;
        const { name } = result;
        this.dispatchCustomEvent(ELEMENT_EVENTS.state.out, {
            state: { name, x: target.left, y: target.top },
            event: e,
        });
        const stroke = name === this.initialState ? __classPrivateFieldGet(this, _initialStateStroke) : __classPrivateFieldGet(this, _stateStroke);
        target._objects[0].set('stroke', stroke.toString());
        target._objects[0].animate('radius', 30, {
            duration: 300,
            onChange: __classPrivateFieldGet(this, _canvas).renderAll.bind(__classPrivateFieldGet(this, _canvas)),
        });
        if (__classPrivateFieldGet(this, _newTransition) == null)
            return;
        const { src, dest, path, pointer, event } = __classPrivateFieldGet(this, _newTransition);
        if (dest) {
            __classPrivateFieldGet(this, _canvas).remove(path);
            __classPrivateFieldGet(this, _canvas).remove(pointer);
            __classPrivateFieldSet(this, _newTransition, this.drawTransition(src.x, src.y, dest.x, dest.y, event));
            __classPrivateFieldGet(this, _newTransition).src = src;
            __classPrivateFieldGet(this, _newTransition).event = event;
            __classPrivateFieldGet(this, _newTransition).dest = null;
        }
    }
    handleStateClick({ button, target: _target, e }) {
        if (_target == null)
            return;
        const target = _target;
        const { name: targetName = '' } = target;
        const result = this.parseObjectName(targetName);
        if (result == null)
            return;
        const { name } = result;
        const detail = {
            state: { name, x: target.left, y: target.top },
            event: e,
        };
        if (button === 3) {
            this.dispatchCustomEvent(ELEMENT_EVENTS.state.rightClick, detail);
        }
        else {
            this.dispatchCustomEvent(ELEMENT_EVENTS.state.click, detail);
        }
        if (__classPrivateFieldGet(this, _newTransition) == null)
            return;
        const { src, dest, event: eventName } = __classPrivateFieldGet(this, _newTransition);
        if (src && dest) {
            this.stateMachineRules[src.name].events[eventName] = dest.name;
            this.transitions.push(__classPrivateFieldGet(this, _newTransition));
            __classPrivateFieldSet(this, _newTransition, null);
            this.handleStateMachineChange();
        }
    }
    handleTransitionClick({ button, target: _target, e }) {
        if (_target == null)
            return;
        const target = _target;
        const { name: targetName = '' } = target;
        const result = this.parseObjectName(targetName);
        if (result == null)
            return;
        const { name } = result;
        const detail = {
            detail: { transtion: { name } },
            event: e,
        };
        if (button === 3) {
            this.dispatchCustomEvent(ELEMENT_EVENTS.transition.rightClick, detail);
        }
        else {
            this.dispatchCustomEvent(ELEMENT_EVENTS.transition.click, detail);
        }
    }
    handleTransitionOver({ target: _target, e: event }) {
        if (_target == null)
            return;
        const target = _target;
        const { name: targetName = '' } = target;
        const result = this.parseObjectName(targetName);
        if (result == null)
            return;
        const { name } = result;
        const detail = {
            detail: { transtion: { name } },
            event: event,
        };
        this.dispatchCustomEvent(ELEMENT_EVENTS.transition.over, detail);
    }
    handleTransitionOut({ target: _target, e: event }) {
        if (_target == null)
            return;
        const target = _target;
        const { name: targetName = '' } = target;
        const result = this.parseObjectName(targetName);
        if (result == null)
            return;
        const { name } = result;
        const detail = {
            detail: { transtion: { name } },
            event: event,
        };
        this.dispatchCustomEvent(ELEMENT_EVENTS.transition.over, detail);
    }
    // DRAW METHODS
    drawTransition(x1, y1, x2, y2, event) {
        const { link, pointer } = this.calcTransitionLink(x1, y1, x2, y2, 30);
        const stroke = colorHash.hex(event);
        const linkPath = new fabric_1.fabric.Path(link, {
            fill: '',
            stroke,
            strokeWidth: 2,
            perPixelTargetFind: true,
            selectable: false,
            hasControls: false,
        });
        linkPath.name = `transition:${event}`;
        linkPath.on('mouseover', this.handleTransitionOver.bind(this));
        linkPath.on('mouseout', this.handleTransitionOut.bind(this));
        linkPath.on('mousedown', this.handleTransitionClick.bind(this));
        const pointerPath = new fabric_1.fabric.Path(pointer, {
            stroke,
            fill: stroke,
            hasControls: false,
            selectable: false,
        });
        __classPrivateFieldGet(this, _canvas).add(linkPath, pointerPath);
        return { path: linkPath, pointer: pointerPath };
    }
    drawState(name, x, y) {
        const stroke = name === this.initialState ? __classPrivateFieldGet(this, _initialStateStroke) : __classPrivateFieldGet(this, _stateStroke);
        const stateCircle = new fabric_1.fabric.Circle({
            originX: 'center',
            originY: 'center',
            fill: __classPrivateFieldGet(this, _stateFill).toString(),
            stroke: stroke.toString(),
            radius: __classPrivateFieldGet(this, _stateRadius),
        });
        const stateText = new fabric_1.fabric.Text(name, {
            originX: 'center',
            originY: 'center',
            fontSize: 10,
            stroke: __classPrivateFieldGet(this, _textStroke).toString(),
        });
        const stateObject = new fabric_1.fabric.Group([stateCircle, stateText], {
            originX: 'center',
            originY: 'center',
            left: x,
            top: y,
            hasControls: false,
            selectable: __classPrivateFieldGet(this, _selectableStates),
            padding: 2,
        });
        stateObject.name = `state:${name}`;
        stateObject.on('mouseout', this.handleStateMouseOut.bind(this));
        stateObject.on('mouseover', this.handleStateMouseOver.bind(this));
        stateObject.on('mousedown', this.handleStateClick.bind(this));
        stateObject.on('mousedblclick', this.handleStateDoubleClick.bind(this));
        __classPrivateFieldGet(this, _canvas).add(stateObject);
        this.states.push(stateObject);
        this.statesMap[name] = stateObject;
        return stateObject;
    }
    // MATH CALCULATIONS
    calcLinkPath(fromX, fromY, toX, toY, curve) {
        const midX = fromX + (toX - fromX) / 2;
        const midY = fromY + (toY - fromY) / 2;
        const pX = midX - curve;
        const pY = midY + curve;
        return `M ${fromX} ${fromY} Q ${pX}, ${pY}, ${toX}, ${toY}`;
    }
    calcPointerPath(x, y, rotation = 0) {
        const height = 10;
        const base = 10;
        const side = Math.sqrt((base / 2) ** 2 + height ** 2);
        const rad = Math.atan(base / 2 / height);
        const l1 = `L ${x + Math.cos(rad + rotation) * side} ${y + Math.sin(rad + rotation) * side}`;
        const l2 = `L ${x + Math.cos(-rad + rotation) * side} ${y + Math.sin(-rad + rotation) * side}`;
        return `M ${x} ${y} ${l1} ${l2} z`;
    }
    calcSelfLinkPath(x, y) {
        const bezierY = y - __classPrivateFieldGet(this, _stateRadius);
        const y1 = bezierY - __classPrivateFieldGet(this, _selfLinkHeight);
        const x2 = x - __classPrivateFieldGet(this, _selfLinkHeight);
        const y2 = bezierY;
        return `M ${x} ${bezierY} C ${x} ${y1}, ${x2} ${y2}, ${x} ${bezierY}`;
    }
    calcTransitionLink(x1, y1, x2, y2, curve) {
        if (x1 === x2 && y1 === y2) {
            return {
                link: this.calcSelfLinkPath(x1, y1),
                pointer: this.calcPointerPath(x1, y1 - __classPrivateFieldGet(this, _stateRadius), -Math.PI / 2),
            };
        }
        const theta = Math.atan((y2 - y1) / (x2 - x1));
        const sign = (theta > 0 ? -1 : 1) * (y2 > y1 ? -1 : 1);
        const fromX = x1 + sign * Math.cos(theta) * __classPrivateFieldGet(this, _stateRadius);
        const fromY = y1 + sign * Math.sin(theta) * __classPrivateFieldGet(this, _stateRadius);
        const toX = x2 + sign * -1 * Math.cos(theta) * __classPrivateFieldGet(this, _stateRadius);
        const toY = y2 + sign * -1 * Math.sin(theta) * __classPrivateFieldGet(this, _stateRadius);
        const midX = fromX + (toX - fromX) / 2;
        const midY = fromY + (toY - fromY) / 2;
        const pX = midX - curve * sign;
        const pY = midY + curve * sign;
        const pointerTheta = Math.atan((toY - pY) / (toX - pX));
        return {
            link: this.calcLinkPath(fromX, fromY, toX, toY, curve * sign),
            pointer: this.calcPointerPath(toX, toY, sign === -1 ? pointerTheta : Math.PI + pointerTheta),
        };
    }
}
exports.default = StateMachineEditor;
_root = new WeakMap(), _canvasElement = new WeakMap(), _canvas = new WeakMap(), _isDragging = new WeakMap(), _newTransition = new WeakMap(), _lastPosX = new WeakMap(), _lastPosY = new WeakMap(), _selectableStates = new WeakMap(), _stateRadius = new WeakMap(), _stateFill = new WeakMap(), _stateStroke = new WeakMap(), _initialStateStroke = new WeakMap(), _textStroke = new WeakMap(), _backgroundColor = new WeakMap(), _selectionColor = new WeakMap(), _selfLinkHeight = new WeakMap(), _maxZoom = new WeakMap(), _minZoom = new WeakMap();
window.customElements.define('state-machine-editor', StateMachineEditor);
// new StateMachineEditor("canvasid", sm)
